from .basemodel import BasePlayer, BaseChatEvent, BaseJoinEvent, BaseQuitEvent


class MineCraftPlayer(BasePlayer):
    """原版 玩家信息"""


class MinecraftPlayerChatEvent(BaseChatEvent):
    """原版 玩家聊天事件"""
    event_name = "MinecraftPlayerChatEvent"
    player: MineCraftPlayer


class MinecraftPlayerJoinEvent(BaseJoinEvent):
    """原版 玩家加入事件"""
    event_name = "MinecraftPlayerJoinEvent"
    player: MineCraftPlayer


class MinecraftPlayerQuitEvent(BaseQuitEvent):
    """原版 玩家退出事件"""
    event_name = "MinecraftPlayerQuitEvent"
    player: MineCraftPlayer



from .basemodel import BasePlayer, BaseChatEvent, BaseDeathEvent, BaseJoinEvent, BaseQuitEvent


class ForgePlayer(BasePlayer):
    """玩家信息"""
    nickname: str
    uuid: str
    ipAddress: str
    level: str
    """地图？"""
    speed: float


class ForgeServerChatEvent(BaseChatEvent):
    """Forge API ServerChatEvent"""
    event_name = "ServerChatEvent"
    player: ForgePlayer


class ForgePlayerLoggedInEvent(BaseJoinEvent):
    """Forge API PlayerLoggedInEvent"""
    event_name = "ForgePlayerLoggedInEvent"
    player: ForgePlayer


class ForgePlayerLoggedOutEvent(BaseQuitEvent):
    """Forge API PlayerLoggedOutEvent"""
    event_name = "ForgePlayerLoggedOutEvent"
    player: ForgePlayer


class ForgePlayerRespawnEvent(BaseDeathEvent):
    """Forge API ForgePlayerRespawnEvent"""
    event_name = "ForgePlayerRespawnEvent"
    player: ForgePlayer



from .basemodel import BaseEvent, BaseChatEvent, BaseDeathEvent, BaseJoinEvent, BaseQuitEvent
from .forge import ForgePlayerLoggedInEvent, ForgePlayerLoggedOutEvent, ForgeServerChatEvent, ForgePlayerRespawnEvent
from .minecraft import MinecraftPlayerChatEvent, MinecraftPlayerJoinEvent, MinecraftPlayerQuitEvent
from .spigot import AsyncPlayerChatEvent, PlayerDeathEvent, PlayerJoinEvent, PlayerQuitEvent

event_dict = {
    # 原版
    "MinecraftPlayerJoinEvent": MinecraftPlayerJoinEvent,
    "MinecraftPlayerQuitEvent": MinecraftPlayerQuitEvent,
    "MinecraftPlayerChatEvent": MinecraftPlayerChatEvent,
    # Spigot
    "AsyncPlayerChatEvent": AsyncPlayerChatEvent,
    "PlayerDeathEvent": PlayerDeathEvent,
    "PlayerJoinEvent": PlayerJoinEvent,
    "PlayerQuitEvent": PlayerQuitEvent,
    # Forge
    "ForgeServerChatEvent": ForgeServerChatEvent,
    "ForgePlayerLoggedInEvent": ForgePlayerLoggedInEvent,
    "ForgePlayerLoggedOutEvent": ForgePlayerLoggedOutEvent,
    "ForgePlayerRespawnEvent": ForgePlayerRespawnEvent,
}



import json
from typing import Union

from nonebot import logger, get_driver
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent
from nonebot.internal.permission import Permission
from nonebot_plugin_guild_patch import GuildMessageEvent

from .config import Config
from .model import event_dict, BaseEvent, BaseChatEvent, BaseDeathEvent, BaseJoinEvent, BaseQuitEvent

plugin_config: Config = Config.parse_obj(get_driver().config)


async def _guild_admin(
        bot: Bot,
        event: GuildMessageEvent,
):
    """检测是否为频道管理员"""
    roles = set(
        role["role_name"]
        for role in (
            await bot.get_guild_member_profile(
                guild_id=event.guild_id, user_id=event.user_id
            )
        )["roles"]
    )
    return bool(roles & set(plugin_config.mc_qq_guild_admin_roles))


GUILD_ADMIN: Permission = Permission(_guild_admin)
"""频道管理员权限"""


async def _msg_to_qq_process(event: BaseEvent) -> str:
    """处理来自MC的消息，并返回处理后的消息"""
    if isinstance(event, BaseChatEvent):
        return f"{event.player.nickname} 说：{event.message}"
    elif isinstance(event, BaseDeathEvent):
        return f"{event.player.nickname} {event.death_message}"
    elif isinstance(event, BaseJoinEvent):
        return f"{event.player.nickname} 加入了游戏"
    elif isinstance(event, BaseQuitEvent):
        return f"{event.player.nickname} 离开了游戏"
    else:
        return "未知消息"


async def _send_msg_to_qq_common(
        bot: Bot,
        message: str,
):
    """发送消息到 QQ"""
    json_msg = json.loads(message)
    event = event_dict[json_msg["event_name"]].parse_obj(json_msg)

    msg = await _msg_to_qq_process(event)
    if plugin_config.mc_qq_display_server_name:
        msg = f"[{event.server_name}] {msg}"
    # 循环服务器列表并发送消息
    if plugin_config.mc_qq_server_list:
        for per_server in plugin_config.mc_qq_server_list:
            # 判断服务器名称是否相同
            if per_server.server_name == event.server_name:
                # 判断是否发送到群聊
                if group_list := per_server.group_list:
                    for per_group in group_list:
                        logger.success(
                            f"[MC_QQ]丨from [{event.server_name}] to [群:{per_group}] \"{msg}\"")
                        await bot.send_group_msg(
                            group_id=per_group,
                            message=msg
                        )
                # 判断是否发送到频道
                if guild_list := per_server.guild_list:
                    for per_guild in guild_list:
                        logger.success(
                            f"[MC_QQ]丨from [{event.server_name}] to [频道:{per_guild.guild_id}/{per_guild.channel_id}] \"{msg}\"")
                        await bot.send_guild_channel_msg(
                            guild_id=per_guild.guild_id,
                            channel_id=per_guild.channel_id,
                            message=msg
                        )


async def get_member_nickname(
        bot: Bot,
        event: Union[GroupMessageEvent, GuildMessageEvent],
        user_id: Union[int, str]
) -> str:
    """获取昵称"""
    # 判断从 群/频道 获取成员信息
    if isinstance(event, GroupMessageEvent):
        # 如果获取发送者的昵称
        if event.user_id == int(user_id):
            # 如果群名片为空，则发送昵称
            return event.sender.card or event.sender.nickname
        # 如果获取其他人的昵称
        else:
            return (await bot.get_group_member_info(
                group_id=event.group_id,
                user_id=user_id,
                no_cache=True
            ))['nickname']
    elif isinstance(event, GuildMessageEvent):
        # 返回频道成员昵称
        if event.user_id == user_id:
            return event.sender.nickname
        else:
            return (await bot.get_guild_member_profile(
                guild_id=event.guild_id,
                user_id=user_id))['nickname']



from aiomcrcon import Client as RconClient
from nonebot.drivers.websockets import WebSocket
from typing import Optional, List, Dict
from pydantic import BaseModel, Extra, Field


class Client:
    """MC_QQ 客户端"""
    server_name: str
    websocket: WebSocket
    rcon: Optional[RconClient] = None

    def __init__(
            self, server_name: str,
            websocket: WebSocket,
            rcon: Optional[RconClient] = None
    ):
        self.server_name: str = server_name
        self.websocket: WebSocket = websocket
        self.rcon: Optional[RconClient] = rcon


CLIENTS: Dict[str, Client] = {}


class Guild(BaseModel):
    """频道配置"""
    # 频道ID
    guild_id: int
    # 子频道ID
    channel_id: int


class Server(BaseModel):
    """服务器配置"""
    # 服务器名称
    server_name: str
    # 服务器群列表
    group_list: Optional[List[int]] = []
    # 服务器频道列表
    guild_list: Optional[List[Guild]] = []
    # 是否开启 Rcon
    rcon_enable: Optional[bool] = False
    # 该群Bot ID
    self_id: Optional[int] = None


class Config(BaseModel, extra=Extra.ignore):
    """配置"""
    # 路由地址
    mc_qq_ws_url: Optional[str] = "/onebot/v11/mcqq"
    # 是否发送群聊名称
    mc_qq_send_group_name: Optional[bool] = False
    # 是否显示服务器名称
    mc_qq_display_server_name: Optional[bool] = False
    # 服务器列表
    mc_qq_server_list: Optional[List[Server]] = Field(default_factory=list)
    # MCRcon 密码
    mc_qq_rcon_password: Optional[str] = "password"
    # Rcon 字典
    mc_qq_rcon_dict: Optional[Dict[str, int]] = Field(default_factory=dict)
    # MC_QQ 频道管理员身份组
    mc_qq_guild_admin_roles: Optional[List[str]] = ["频道主", "超级管理员"]



import contextlib
from aiomcrcon import Client as RconClient
from nonebot import logger, get_bot
from nonebot.drivers import URL, ReverseDriver, WebSocketServerSetup
from nonebot.drivers.websockets import WebSocket, WebSocketClosed

from .common import plugin_config
from .config import Client, CLIENTS
from .utils import _send_msg_to_qq, _rcon_connect


async def _ws_handler(websocket: WebSocket):
    """WebSocket"""
    try:
        server_name = websocket.request.headers.get("x-self-name").encode('utf-8').decode('unicode_escape')
    except KeyError as e:
        # 服务器名为空
        logger.error(f"[MC_QQ]丨未获取到该服务器的名称，连接断开：{e}")
        await websocket.close(1008, "[MC_QQ]丨未获取到该服务器的名称，连接断开")
        return
    else:

        if CLIENTS.get(server_name):
            # 服务器名已存在
            logger.error(f"[MC_QQ]丨已有相同服务器名的连接，连接断开")
            await websocket.close(1008, "[MC_QQ]丨已有相同服务器名的连接")
            return

        rcon_client = None
        bot_self_id = None
        for server in plugin_config.mc_qq_server_list:
            if server_name == server.server_name:
                if server.rcon_enable:
                    rcon_client = RconClient(
                        websocket.__dict__["websocket"].__dict__["scope"]["client"][0],
                        plugin_config.mc_qq_rcon_dict[server_name],
                        plugin_config.mc_qq_rcon_password
                    )
                    await _rcon_connect(rcon_client=rcon_client, server_name=server_name)
                    bot_self_id = str(server.self_id) if server.self_id else None
                    break
        CLIENTS[server_name] = Client(
            server_name=server_name,
            websocket=websocket,
            rcon=rcon_client
        )

        await websocket.accept()

        logger.success(f"[MC_QQ]丨[Server:{server_name}] 已连接至 WebSocket 服务器")

        try:
            while True:
                try:
                    message = await websocket.receive()
                    # 获取指定ID Bot
                    bot = get_bot(bot_self_id)
                except KeyError as e:
                    logger.warning(f"[MC_QQ]丨[Server:{server_name}] 对应 self_id 的 Bot 不存在：{e}")
                except ValueError as e:
                    logger.warning(f"[MC_QQ]丨[Server:{server_name}] 未指定Bot，且当前无其他Bot可用：{e}")
                else:
                    # 以指定ID Bot 发送消息
                    await _send_msg_to_qq(bot=bot, message=message)

        except WebSocketClosed as e:
            logger.warning(f"[MC_QQ]丨[Server:{server_name}] 的 WebSocket 出现异常：{e}")
        finally:
            with contextlib.suppress(Exception):
                await websocket.close()
            if rcon_client:
                await rcon_client.close()
            del CLIENTS[server_name]


async def set_route(driver: ReverseDriver):
    driver.setup_websocket_server(
        WebSocketServerSetup(
            path=URL("/onebot/v11/mcqq"),
            name="mcqq",
            handle_func=_ws_handler,
        )
    )



import json
import contextlib
from typing import List, Union, Tuple

from aiomcrcon import (
    Client as RconClient,
    ClientNotConnectedError,
    RCONConnectionError,
    IncorrectPasswordError
)
from nonebot import logger
from nonebot.adapters.onebot.v11 import Bot, GroupMessageEvent
from nonebot.drivers.websockets import WebSocketClosed
from nonebot_plugin_guild_patch import GuildMessageEvent

from .common import get_member_nickname, _send_msg_to_qq_common, plugin_config
from .config import Client, CLIENTS


async def send_msg_to_mc(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent]):
    """
    发送消息到 MC
    :param bot: Bot
    :param event: 事件
    """
    # 处理来自QQ的消息
    if client_list := await _get_clients(event=event):
        for client in client_list:
            if client:
                # 先判断是否有Rcon进行发送
                if client.rcon:
                    rcon_text_msg, msg_cmd = await _msg_process_to_cmd(bot=bot, event=event)
                    try:
                        await client.rcon.send_cmd(msg_cmd)
                        logger.success(f"[MC_QQ_Rcon]丨发送至 [server:{client.server_name}] 的消息 \"{rcon_text_msg}\"")
                    except ClientNotConnectedError as e:
                        logger.error(f"[MC_QQ_Rcon]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：{e}")
                        await remove_client(client.server_name)
                elif client.websocket:
                    text_msg, msg_json = await _msg_process_to_json(bot=bot, event=event)
                    try:
                        await client.websocket.send(msg_json)
                        logger.success(f"[MC_QQ]丨发送至 [server:{client.server_name}] 的消息 \"{text_msg}\"")
                    except WebSocketClosed as e:
                        logger.error(f"[MC_QQ]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：{e}")
                        await remove_client(client.server_name)
                else:
                    logger.error(f"[MC_QQ]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：该服务器没有连接")
                    await remove_client(client.server_name)


async def send_cmd_to_mc(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent], cmd: str):
    """
    发送命令到 MC
    :param bot: Bot
    :param event: 事件
    :param cmd: 命令
    """
    # 处理来自QQ的消息
    if client_list := await _get_clients(event=event):
        for client in client_list:
            if client:
                # 先判断是否有Rcon进行发送
                if client.rcon:
                    try:
                        back_msg = await client.rcon.send_cmd(cmd)
                        await bot.send(event=event, message=f"服务器返回：{back_msg[0]}")
                        logger.success(f"[MC_QQ_Rcon]丨发送至 [server:{client.server_name}] 的消息 \"{cmd}\"")
                    except ClientNotConnectedError as e:
                        logger.error(f"[MC_QQ_Rcon]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：{e}")
                        await remove_client(client.server_name)
                elif client.websocket:
                    cmd_list = {"message": [{"msgType": "command", "msgData": cmd}]}
                    await client.websocket.send_text(str(cmd_list))
                else:
                    logger.error(f"[MC_QQ]丨发送至 [Server:{client.server_name}] 的过程中出现了错误：该客户端没有连接")
                    await remove_client(client.server_name)


async def _get_clients(event: Union[GroupMessageEvent, GuildMessageEvent]) -> List[Client]:
    """
    获取 服务器名、ws客户端, 返回client列表
    :param event: 事件
    :return: client列表
    """
    res: List[Client] = []
    for per_server in plugin_config.mc_qq_server_list:
        if isinstance(event, GroupMessageEvent):
            if event.group_id in per_server.group_list:
                res.append(CLIENTS.get(per_server.server_name))
        if isinstance(event, GuildMessageEvent):
            for per_guild in per_server.guild_list:
                if per_guild.guild_id == event.guild_id and per_guild.channel_id == event.channel_id:
                    res.append(CLIENTS.get(per_server.server_name))
    return res


async def remove_client(server_name: str):
    """
    移除客户端
    :param server_name: 服务器名
    """
    if client := CLIENTS.get(server_name):
        if client.websocket:
            with contextlib.suppress(Exception):
                await client.websocket.close()
        if client.rcon:
            await client.rcon.close()
        del CLIENTS[server_name]


async def rcon_connect(rcon_client: RconClient, server_name: str):
    """
    连接 Rcon
    :param rcon_client: Rcon 客户端
    :param server_name: 服务器名
    """
    try:
        await rcon_client.connect()
        logger.success(f"[MC_QQ]丨[Server:{server_name}] 的Rcon连接成功")
    except RCONConnectionError as e:
        logger.error(f"[MC_QQ]丨[Server:{server_name}] 的Rcon连接失败：{str(e)}")
    except IncorrectPasswordError as e:
        logger.error(f"[MC_QQ]丨[Server:{server_name}] 的Rcon密码错误：{str(e)}")


async def send_msg_to_qq(
        bot: Bot,
        message: str
):
    """
    发送消息到 QQ
    :param bot: Bot
    :param message: 消息
    """
    await _send_msg_to_qq_common(
        bot=bot,
        message=message
    )


async def _msg_process_to_json(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent]) -> Tuple[str, str]:
    """
    消息处理为 JSON
    :param bot: Bot
    :param event: 事件
    :return: text_msg, msgJson
    """
    # 获取昵称
    member_nickname = await get_member_nickname(bot, event, event.user_id)

    # 初始化消息
    text_msg = member_nickname + "说："

    # 初始化消息字典
    message_list = []

    # 发送群聊名称
    if plugin_config.mc_qq_send_group_name:
        group_name = {'msgType': "group_name"}
        if isinstance(event, GroupMessageEvent):
            group_name['msgData'] = (await bot.get_group_info(group_id=event.group_id))['group_name']
        elif isinstance(event, GuildMessageEvent):
            guild_name = (await bot.get_guild_meta_by_guest(guild_id=event.guild_id))['guild_name']
            for per_channel in (await bot.get_guild_channel_list(guild_id=event.guild_id, no_cache=True)):
                if str(event.channel_id) == per_channel['channel_id']:
                    channel_name = per_channel['channel_name']
                    group_name['msgData'] = f"{guild_name}丨{channel_name}"
                    break
        message_list.append({"msgType": "group_name", "msgData": group_name})

    # 将群成员昵称装入消息列表
    message_list.append({"msgType": "senderName", "msgData": member_nickname})

    for msg in event.message:
        per_msg = {'msgType': msg.type}
        # 文本
        if msg.type == "text":
            msgData = msg.data['text'].replace("\r", "").replace("\n", "\n * ")
            text_msg += msgData
        # 图片
        elif msg.type == "image":
            msgData = msg.data['url']
            text_msg += '[图片]'
        # 表情
        elif msg.type == "face":
            msgData = '[表情]'
            text_msg += '[表情]'
        # 语音
        elif msg.type == "record":
            msgData = '[语音]'
            text_msg += '[语音]'
        # 视频
        elif msg.type == "video":
            msgData = msg.data['url']
            text_msg += '[视频]'
        # @
        elif msg.type == "at":
            # 获取被@ 群/频道 昵称
            at_member_nickname = await get_member_nickname(bot, event, msg.data['qq'])
            msgData = f"@{at_member_nickname}"
            text_msg += msgData
        # share
        elif msg.type == "share":
            msgData = msg.data['url']
            text_msg += '[分享：' + msg.data['title'] + ']'
        # forward
        elif msg.type == "forward":
            # TODO 将合并转发消息拼接为字符串
            # 获取合并转发 await bot.get_forward_msg(message_id=event.message_id)
            msgData = '[合并转发]'
            text_msg = msgData
        else:
            msgData = msg.type
            text_msg += '[' + msg.type + '] '

        text_msg += " "

        # 装入消息数据
        per_msg['msgData'] = msgData
        # 放入消息列表
        message_list.append(per_msg)

    return text_msg, str({"message": message_list})


async def _msg_process_to_cmd(bot: Bot, event: Union[GroupMessageEvent, GuildMessageEvent]) -> Tuple[str, str]:
    """
    消息处理为 命令
    :param bot: Bot
    :param event: 事件
    :return: text_msg, command_msg
    """
    # 获取昵称
    member_nickname = await get_member_nickname(bot, event, event.user_id)

    # 初始化日志消息
    text_msg = member_nickname + " 说："

    command_msg = "tellraw @a "

    message_list = [
        {"text": "[MC_QQ] ", "color": "yellow"},
    ]
    if plugin_config.mc_qq_send_group_name:
        if isinstance(event, GroupMessageEvent):
            message_list.append(
                {"text": (await bot.get_group_info(group_id=event.group_id))['group_name'] + " ", "color": "aqua"})
        elif isinstance(event, GuildMessageEvent):
            guild_name = (await bot.get_guild_meta_by_guest(guild_id=event.guild_id))['guild_name']
            for per_channel in (await bot.get_guild_channel_list(guild_id=event.guild_id, no_cache=True)):
                if str(event.channel_id) == per_channel['channel_id']:
                    message_list.append(
                        {"text": guild_name + "丨" + per_channel['channel_name'] + " ", "color": "aqua"})
                    break
    message_list.append({"text": member_nickname, "color": "aqua"})
    message_list.append({"text": " 说：", "color": "yellow"})

    for msg in event.message:
        # 文本
        if msg.type == "text":
            msg_dict = {"text": msg.data['text'].replace("\r", "").replace("\n", "\n * ") + " ", "color": "white"}
            text_msg += msg.data['text'].replace("\r", "").replace("\n", "\n * ")
        # 图片
        elif msg.type == "image":
            msg_dict = {"text": "[图片] ", "color": "yellow",
                        "clickEvent": {"action": "open_url", "value": msg.data['url']},
                        "hoverEvent": {"action": "show_text", "contents": [{"text": "查看图片", "color": "gold"}]}
                        }
            text_msg += '[图片]'
        # 表情
        elif msg.type == "face":
            msg_dict = {"text": "[表情] ", "color": "gold"}
            text_msg += '[表情]'
        # 语音
        elif msg.type == "record":
            msg_dict = {"text": "[语音] ", "color": "light_purple"}
            text_msg += '[语音]'
        # 视频
        elif msg.type == "video":
            msg_dict = {"text": "[视频] ", "color": "light_purple",
                        "clickEvent": {"action": "open_url", "value": msg.data['url']},
                        "hoverEvent": {"action": "show_text",
                                       "contents": [{"text": "查看视频", "color": "dark_purple"}]}
                        }
            text_msg += '[视频]'
        # @
        elif msg.type == "at":
            # 获取被@ 群/频道 昵称
            at_member_nickname = await get_member_nickname(bot, event, msg.data['qq'])
            msg_dict = {"text": "@" + at_member_nickname + " ", "color": "green"}
            text_msg += f"@{at_member_nickname}"
        # share
        elif msg.type == "share":
            msg_dict = {"text": "[分享：" + msg.data['title'] + "] ", "color": "yellow",
                        "clickEvent": {"action": "open_url", "value": msg.data['url']},
                        "hoverEvent": {"action": "show_text", "contents": [{"text": "查看图片", "color": "gold"}]}
                        }
            text_msg += '[分享：' + msg.data['title'] + ']'
        # forward
        elif msg.type == "forward":
            # TODO 将合并转发消息拼接为字符串
            # 获取合并转发 await bot.get_forward_msg(message_id=event.message_id)
            msg_dict = {"text": "[合并转发] ", "color": "white"}
            text_msg += '[合并转发]'
        else:
            msg_dict = {"text": "[ " + msg.type + "] ", "color": "white"}
            text_msg += '[' + msg.type + ']'

        # 放入消息列表
        message_list.append(msg_dict)

    # 拼接完整命令
    command_msg += json.dumps(message_list)
    return text_msg, command_msg
